/*
 *  id3proc.c
 *
 *  ID3 tag editor and viewer
 *
 *  version 0.1 (22.10.2002)
 *  by Michal Miszewski <fixer@irc.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#define VERSION "0.1"

FILE *fd;

int allhdr = 0, html = 0, nohdrs = 0, edit = 0;

/* defaults */
char defsong[31] = "", defartist[31] = "", defalbum[31] = "",
  defyear[5] = "", defcomment[31] = "";
unsigned char defgenre = 127;

struct id3
{
  char tag[3], song[30], artist[30], album[30],
    year[4], comment[30];
  unsigned char genre;
};

/* converts genre number into a name */
void genre(int genre, char buf[32])
{
  char genres[116][28] = {
  "Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk", "Grunge",
  "Hip-Hop", "Jazz", "Metal", "New Age", "Oldies", "Other", "Pop", "R&B",
  "Rap", "Reggae", "Rock", "Techno", "Industrial", "Alternative", "Ska",
  "Death Metal", "Pranks", "Soundtrack", "Euro-Techno", "Ambient", "Trip-Hop",
  "Vocal", "Jazz+Funk", "Fusion", "Trance", "Classical", "Instrumental",
  "Acid", "House", "Game", "Sound Clip", "Gospel", "Noise", "AlternRock",
  "Bass", "Soul", "Punk", "Space", "Meditative", "Instrumental Pop",
  "Instrumental Rock", "Ethnic", "Gothic", "Darkwave", "Techno-Industrial",
  "Electronic", "Pop-Folk", "Eurodance", "Dream", "Southern Rock", "Comedy",
  "Cult", "Gangsta", "Top 40", "Christian Rap", "Pop/Funk", "Jungle",
  "Native American", "Cabaret", "New Wave", "Psychadelic", "Rave", "Showtunes",
  "Trailer", "Lo-Fi", "Tribal", "Acid Punk", "Acid Jazz", "Polka", "Retro",
  "Musical", "Rock & Roll", "Hard Rock", "Folk", "Folk/Rock",
  "National Folk", "Swing", "Fast-fusion", "Bebob", "Latin", "Revival",
  "Celtic", "Bluegrass", "Avantgarde", "Gothic Rock", "Progressive Rock",
  "Psychedelic Rock", "Symphonic Rock", "Slow Rock", "Big Band", "Chorus",
  "Easy Listening", "Acoustic", "Humour", "Speech", "Chanson", "Opera",
  "Chamber Music", "Sonata", "Symphony", "Booty Bass", "Primus", "Porn Groove",
  "Satire", "Slow Jam", "Club", "Tango", "Samba", "Folklore"};

  if (genre < 116) strcpy(buf, genres[genre]); else strcpy(buf, "");
}

/* prints a list of genres */
void print_genres(void)
{
  char buf[50];
  int a;

  for (a = 0; a < 116; a++)
  {
    genre(a, buf);
    fprintf(stdout, "(%d)%s  ", a, buf);
  }
  fprintf(stdout, "\n");
}

/* converts tag record into a string */
void to_str(char in[30], char buf[32], int size)
{
  int len, a, b;

  len = size - 1;
  for (a = 0; a < len; a++)
    if ((in[len - a] != ' ') && (in[len - a] != 0) && (in[len - a] != 20)) break;
  for (b = 0; b < (len - a + 1); b++) buf[b] = in[b];
  buf[b] = '\0';
  if (strcmp(buf, " ") == 0) strcpy(buf, "");
}

/* converts a string into a record */
void to_tagrecord(char in[31], char out[], int size)
{
  int a;

  for (a = 0; a < size; a++) out[a] = 0;
  for (a = 0; a < strlen(in); a++)
    if ((a < size) && (in[a] != '\n')) out[a] = in[a]; else break;
}

/* text input */
void get_name(char def[], char back[])
{
  char buf[200];
  fgets(buf, 200, stdin);
  if (strcmp(buf, "\n") == 0) strcpy(buf, def);
  strncpy(back, buf, 31);
}

/* checks if there is any text in a tag record */
int is_text(char buf[32], int size)
{
  int a;

  for (a = 0; a < (size - 1); a++)
    if ((buf[a] != ' ') && (buf[a] != 0)) return 1;
  return 0;
}

/* tag edition by user */
void edit_tag(struct id3 *tag, char fname[])
{
  char buf[32], back[32];
  char tmp;
  int ret = 1;

  fprintf(stdout, "File: %s\n", fname);
  if (strcmp(defsong, "") != 0) strcpy(buf, defsong);
  else to_str(tag->song, buf, 30);
  fprintf(stdout, "Song [%s]: ", buf);
  get_name(buf, back); to_tagrecord(back, tag->song, 30);
  if (strcmp(defartist, "") != 0) strcpy(buf, defartist);
  else to_str(tag->artist, buf, 30);
  fprintf(stdout, "Artist [%s]: ", buf);
  get_name(buf, back); to_tagrecord(back, tag->artist, 30);
  if (strcmp(defalbum, "") != 0) strcpy(buf, defalbum);
  else to_str(tag->album, buf, 30);
  fprintf(stdout, "Album [%s]: ", buf);
  get_name(buf, back); to_tagrecord(back, tag->album, 30);
  if (strcmp(defyear, "") != 0) strcpy(buf, defyear);
  else to_str(tag->year, buf, 4);
  fprintf(stdout, "Year [%s]: ", buf);
  get_name(buf, back); to_tagrecord(back, tag->year, 4);
  if (strcmp(defcomment, "") != 0) strcpy(buf, defcomment);
  else to_str(tag->comment, buf, 30);
  fprintf(stdout, "Comment [%s]: ", buf);
  get_name(buf, back); to_tagrecord(back, tag->comment, 30);
  while(ret)
  {
    ret = 0;
    if (defgenre < 116) sprintf(buf, "%d", defgenre);
    else sprintf(buf, "%d", tag->genre);
    fprintf(stdout, "Genre [%s, ? for list]: ", buf);
    get_name(buf, back);
    if (back[0] == '?')
    {
      print_genres();
      ret = 1;
    }
    else
    {
      tmp = atoi(back);
      if (tmp < 116) tag->genre = tmp; else tag->genre = 12;
    }
  }
  fprintf(stdout, "\n");
}

/* prints the contents of an ID3 tag */
void print_tag(struct id3 tag, char fname[])
{
  char buf[32] = "";

  fprintf(stdout, "File:     %s\n", fname);
  to_str(tag.song, buf, 30);
  if ((strcmp(buf, "") != 0) || allhdr) fprintf(stdout, "Song:     %s\n", buf);
  to_str(tag.artist, buf, 30);
  if ((strcmp(buf, "") != 0) || allhdr) fprintf(stdout, "Artist:   %s\n", buf);
  to_str(tag.album, buf, 30);
  if ((strcmp(buf, "") != 0) || allhdr) fprintf(stdout, "Album:    %s\n", buf);
  to_str(tag.year, buf, 4);
  if ((strcmp(buf, "") != 0) || allhdr) fprintf(stdout, "Year:     %s\n", buf);
  to_str(tag.comment, buf, 30);
  if ((strcmp(buf, "") != 0) || allhdr) fprintf(stdout, "Comment:  %s\n", buf);
  if ((tag.genre < 116) || allhdr)
  {
    genre(tag.genre, buf);
    fprintf(stdout, "Genre:    %s\n", buf);
  }
  fprintf(stdout, "\n");
}

/* prints the contents of an ID3 tag as HTML code */
void print_tag_html(struct id3 tag, char fname[])
{
  char buf[32];
  fprintf(stdout, "<TR>");
  fprintf(stdout, "<TD><A href=\"%s\">%s</A></TD>", fname, fname);
  fprintf(stdout, "<TD>");
  to_str(tag.song, buf, 30);
  if (strcmp(buf, "") != 0) fprintf(stdout, "%s</TD>", buf);
  else fprintf(stdout, "-</TD>");
  fprintf(stdout, "<TD>");
  to_str(tag.artist, buf, 30);
  if (strcmp(buf, "") != 0) fprintf(stdout, "%s</TD>", buf);
  else fprintf(stdout, "-</TD>");
  fprintf(stdout, "<TD>");
  to_str(tag.album, buf, 30);
  if (strcmp(buf, "") != 0) fprintf(stdout, "%s</TD>", buf);
  else fprintf(stdout, "-</TD>");
  fprintf(stdout, "<TD>");
  to_str(tag.year, buf, 4);
  if (strcmp(buf, "") != 0) fprintf(stdout, "%s</TD>", buf);
  else fprintf(stdout, "-</TD>");
  fprintf(stdout, "<TD>");
  to_str(tag.comment, buf, 30);
  if (strcmp(buf, "") != 0) fprintf(stdout, "%s</TD>", buf);
  else fprintf(stdout, "-</TD>");
  fprintf(stdout, "<TD>");
  if (tag.genre < 116)
  {
    genre(tag.genre, buf);
    fprintf(stdout, "%s</TD>", buf);
  } else fprintf(stdout, "-</TD>");
  fprintf(stdout, "</TR>\n");
}

/* header of a HTML file */
void html_head(void)
{
  fprintf(stdout,
  "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
  "<HTML>\n<HEAD>\n"
  "<TITLE>MP3 files</TITLE>\n"
  "<META http-equiv=\"Content-type\" content=\"text/html; charset=iso-8859-1\">\n"
  "<META http-equiv=\"Content-Language\" content=\"en\">\n"
  "</HEAD>\n"
  "<BODY>\n\n<TABLE border=\"1\">\n"
  "<TR><TH>File</TH><TH>Song</TH><TH>Artist</TH><TH>Album</TH><TH>Year"
  "</TH><TH>Comment</TH><TH>Genre</TH></TR>\n");
}

/* end of a HTML file */
void html_tail(void)
{
  fprintf(stdout, "</TABLE>\n\n</BODY>\n</HTML>\n");
}

/* clears an id3 structure */
void clear_tag(struct id3 *tag)
{
  strcpy(tag->tag, "");
  strcpy(tag->song, "");
  strcpy(tag->artist, "");
  strcpy(tag->album, "");
  strcpy(tag->year, "");
  strcpy(tag->comment, "");
  tag->genre = 127;
}

/* reads an ID3 tag from a file */
int read_tag(struct id3 *tag)
{
  struct stat fbuf;
  char buf[4] = "    ";
  int size;

  fstat(fileno(fd), &fbuf);
  size = (int)fbuf.st_size;
  if (size > 128)
  {
    fseek(fd, (size - 128), 0);
    if (fread(&*tag, sizeof(*tag), 1, fd) != 1) perror("read_tag:fread");
    to_str(tag->tag, buf, 3);
    if (strncmp(buf, "TAG", 3) == 0) return 1; else return 0;
  } else return 0;
}

/* writes an ID3 tag to a file */
void write_tag(struct id3 tag, int is)
{
  struct stat fbuf;

  fstat(fileno(fd), &fbuf);
  if (is) fseek(fd, ((int)fbuf.st_size - 128), 0);
  else fseek(fd, ((int)fbuf.st_size), 0);
  if (fwrite(&tag, sizeof(tag), 1, fd) != 1) perror("write_tag:fwrite");
}

/* opens an mp3 file */
int open_file(char name[])
{
  fd = fopen(name, "r+");
  if (fd == NULL) return 0; else return 1;
}

void process_file(char name[])
{
  struct id3 tag;
  int is = 1;

  if (open_file(name))
  {
    clear_tag(&tag);
    if (!read_tag(&tag))
    {
      clear_tag(&tag);
      is = 0;
    }
    if (edit)
    {
      tag.tag[0] = 'T';
      tag.tag[1] = 'A';
      tag.tag[2] = 'G';
      edit_tag(&tag, name);
      write_tag(tag, is);
    }
    else
    {
      if (html) print_tag_html(tag, name);
      else print_tag(tag, name);
    }
    fclose(fd);
  } else perror(name);
}

/* prints usage instructions */
void usage(char name[])
{
  fprintf(stderr, "ID3 tag editor and viewer v.%s by Michal Miszewski <fixer@irc.pl>\n"
  "THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK!\n\n"
  "usage: %s [-hatndesrlycg] file(s)\n\noptions:\n"
  "    -h\t\tdisplay this help screen\n"
  "    -a\t\twhen listing, display labels for empty fields too\n"
  "    -t\t\tHTML listing\n"
  "    -n\t\tskip HTML file headers\n"
  "    -d\t\tdisplay genre numbers list\n"
  "    -e\t\tedit tag fields\n"
  "    -s text\tdefault name of a song\n"
  "    -r text\tdefault name of an artist\n"
  "    -l text\tdefault name of an album\n"
  "    -y text\tdefault year of edition\n"
  "    -c text\tdefault file comment\n"
  "    -g num\tdefault genre (a number, -d to see a list)\n", VERSION, name);
  exit(EXIT_FAILURE);
}

int main(int argc, char **argv)
{
  int ch;

  while(1)
  {
    ch = getopt(argc, argv, "hatndes:r:l:y:c:g:");
    if (ch == -1) break;
    switch (ch)
    {
      case 'h': { usage(argv[0]); break; }
      case 'a': { allhdr = 1; break; }
      case 't': { html = 1; break; }
      case 'n': { nohdrs = 1; break; }
      case 'd': { print_genres(); exit(EXIT_FAILURE); break; }
      case 'e': { edit = 1; break; }
      case 's': { if (optarg) strncpy(defsong, optarg, 30); break; }
      case 'r': { if (optarg) strncpy(defartist, optarg, 30); break; }
      case 'l': { if (optarg) strncpy(defalbum, optarg, 30); break; }
      case 'y': { if (optarg) strncpy(defyear, optarg, 4); break; }
      case 'c': { if (optarg) strncpy(defcomment, optarg, 30); break; }
      case 'g': { if (optarg) defgenre = atoi(optarg); break; }
    }
  }
  if (optind < argc)
  {
    if ((html) && (!nohdrs)) html_head();
    while (optind < argc) process_file(argv[optind++]);
    if ((html) && (!nohdrs)) html_tail();
  } else fprintf(stderr, "%s: must supply a file name\n", argv[0]);
  return 0;
}
